# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import dataclasses
import re
import textwrap
import typing
from enum import Enum

import parsley


@dataclasses.dataclass
class Term:
    pass


class FilterOn(Enum):
    name = 'name'
    summary = 'summary'
    name_or_summary = 'name_or_summary'


@dataclasses.dataclass
class Filter(Term):
    filter_on: FilterOn
    value: str


@dataclasses.dataclass
class And(Term):
    lhs: Term
    rhs: Term


@dataclasses.dataclass
class Or(Term):
    lhs: Term
    rhs: Term


@dataclasses.dataclass
class Not(Term):
    term: Term


def normalise_name(name: str) -> str:
    return re.sub(r"[-_.]+", "-", name).lower()


# A safe SQL statement must not use *any* user-defined input in the resulting first argument (the SQL query),
# rather any user input MUST be provided as part of the arguments (second part of the value), which will be passed
# to SQLITE to deal with.
SafeSQLStmt = typing.Tuple[str, typing.Tuple[typing.Any, ...]]


def prepare_name(term: Filter) -> SafeSQLStmt:
    if term.value.startswith('"'):
        # Match the phase precisely.
        value = term.value[1:-1]
    else:
        value = normalise_name(term.value)
    value = value.replace('*', '%')
    return "canonical_name LIKE ?", (f'%{value}%',)


def prepare_summary(term: Filter) -> SafeSQLStmt:
    if term.value.startswith('"'):
        # Match the phase precisely.
        value = term.value[1:-1]
    else:
        value = term.value
    value = value.replace('*', '%')
    return "summary LIKE ?", (f'%{value}%',)


def build_sql(term: typing.Union[Term, typing.Tuple[Term, ...]]) -> SafeSQLStmt:
    # Return query and params to be used in SQL. query MUST not be produced using untrusted input, as is vulnerable to SQL injection.
    # Instead, any user input must be in the parameters, which undergoes sqllite built-in cleaning.
    if isinstance(term, tuple):
        if len(term) == 0:
            return '', ()

        # No known query can produce a multi-value term
        assert len(term) == 1
        return build_sql(term[0])

    if isinstance(term, Filter):
        if term.filter_on == FilterOn.name_or_summary:
            sql1, terms1 = prepare_name(term)
            sql2, terms2 = prepare_summary(term)
            return f"({sql1} OR {sql2})", terms1 + terms2
        elif term.filter_on == FilterOn.name:
            return prepare_name(term)
        elif term.filter_on == FilterOn.summary:
            return prepare_summary(term)
        else:
            raise ValueError(f"Unhandled filter on {term.filter_on}")
    elif isinstance(term, And):
        sql1, terms1 = build_sql(term.lhs)
        sql2, terms2 = build_sql(term.rhs)
        return f"({sql1} AND {sql2})", terms1 + terms2
    elif isinstance(term, Or):
        sql1, terms1 = build_sql(term.lhs)
        sql2, terms2 = build_sql(term.rhs)
        return f"({sql1} OR {sql2})", terms1 + terms2
    elif isinstance(term, Not):
        sql1, terms1 = build_sql(term.term)
        return f'(Not {sql1})', terms1
    else:
        raise ValueError(f"unknown term type {type(term)}")


def query_to_sql(query) -> SafeSQLStmt:
    terms = parse(query)
    return build_sql(terms)


grammar = parsley.makeGrammar(
    textwrap.dedent("""\
    quoted_string = '"' <(~'"' anything)*>:c '"' -> f'"{c}"'

    # Anything that looks like a Python package name. Note that wildcard (*) is supported.
    name_like = <letter (letterOrDigit | '-' | '_' | '.' | '*')*>

    # Allow specifying specific fields to filter on
    filter_on = (
        'name:' -> FilterOn.name
        |'summary:' -> FilterOn.summary
        | -> FilterOn.name_or_summary
    )

    # Either a quoted string, or a name-like
    filter_value = (quoted_string | name_like):term -> term

    filter = filter_on:filter_on filter_value:filter_value -> Filter(filter_on, filter_value)

    AND_OR = (ws 'AND' ws -> And
             |ws 'OR' ws -> Or
             |ws -> And)
    filters = (
            filters:lhs AND_OR:op filters:rhs -> op(lhs, rhs)
            |'(' ws? filters:filters ws? ')' -> filters
            |filter:filter -> filter
            |'-' filters:filters -> Not(filters)
    )
    search_terms = (filters+:filters -> tuple(filters)
                   | -> ())
    """),
    {
        'And': And,
        'Or': Or,
        'Filter': Filter,
        'Not': Not,
        'FilterOn': FilterOn,
    },
)


def parse(query: str) -> typing.Tuple[Term, ...]:
    return grammar(query.strip()).search_terms()


ParseError = parsley.ParseError


def simple_name_from_query(terms: typing.Tuple[Term, ...]) -> typing.Optional[str]:
    """If possible, give a simple (normalized) package name which represents the query terms provided"""
    for term in terms:
        if isinstance(term, Filter):
            if term.filter_on in [FilterOn.name_or_summary, FilterOn.name]:
                if '*' in term.value or '"' in term.value:
                    break
                return normalise_name(term.value)
        else:
            break
    return None



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import logging
import sqlite3
import typing
from pathlib import Path

import aiosqlite
import diskcache
import fastapi
import httpx
from simple_repository import SimpleRepository
from simple_repository.components.http import HttpRepository

from . import controller, crawler, errors, fetch_projects, model, view
from .metadata_injector import MetadataInjector


class AppBuilder:
    def __init__(
        self,
        url_prefix: str,
        index_url: str,
        cache_dir: Path,
        template_paths: typing.Sequence[Path],
        static_files_path: Path,
        crawl_popular_projects: bool,
        browser_version: str,
    ) -> None:
        self.url_prefix = url_prefix
        self.index_url = index_url
        self.cache_dir = cache_dir
        self.template_paths = template_paths
        self.static_files_path = static_files_path
        self.crawl_popular_projects = crawl_popular_projects
        self.browser_version = browser_version

        self.cache = diskcache.Cache(str(cache_dir/'diskcache'))
        self.db_path = cache_dir / 'projects.sqlite'
        self.con = sqlite3.connect(
            self.db_path,
            detect_types=sqlite3.PARSE_DECLTYPES,
            timeout=5,
        )
        self.con.row_factory = sqlite3.Row
        fetch_projects.create_table(self.con)

    def create_app(self) -> fastapi.FastAPI:
        _view = self.create_view()

        async def lifespan(app: fastapi.FastAPI):
            async with (
                httpx.AsyncClient() as http_client,
                aiosqlite.connect(self.db_path, timeout=5) as db,
            ):
                _controller = self.create_controller(
                    model=self.create_model(
                        http_client=http_client,
                        database=db,
                    ),
                    view=_view,
                )
                router = _controller.create_router(self.static_files_path)
                app.mount(self.url_prefix or "/", router)
                yield

        app = fastapi.FastAPI(
            lifespan=lifespan,
        )

        # TODO: refactor into a controller
        async def catch_exceptions_middleware(request: fastapi.Request, call_next):
            try:
                return await call_next(request)
            except errors.RequestError as err:
                status_code = err.status_code
                detail = err.detail
            except Exception as err:
                status_code = 500
                detail = f"Internal server error ({err})"
                # raise
                logging.error(err)
            content = _view.error_page(
                request=request,
                context=model.ErrorModel(detail=detail),
            )
            return fastapi.responses.HTMLResponse(
                content=content,
                status_code=status_code,
            )

        app.middleware('http')(catch_exceptions_middleware)

        return app

    def create_view(self) -> view.View:
        return view.View(self.template_paths, self.browser_version)

    def create_crawler(self, http_client: httpx.AsyncClient, source: SimpleRepository) -> crawler.Crawler:
        return crawler.Crawler(
            http_client=http_client,
            crawl_popular_projects=self.crawl_popular_projects,
            source=source,
            projects_db=self.con,
            cache=self.cache,
        )

    def create_model(self, http_client: httpx.AsyncClient, database: aiosqlite.Connection) -> model.Model:
        source = MetadataInjector(
            HttpRepository(
                url=self.index_url,
                http_client=http_client,
            ),
            database=database,
            http_client=http_client,
        )
        return model.Model(
            source=source,
            projects_db=self.con,
            cache=self.cache,
            crawler=self.create_crawler(http_client, source),
        )

    def create_controller(self, view: view.View, model: model.Model) -> controller.Controller:
        return controller.Controller(
            model=model,
            view=view,
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import asyncio
import logging
import os
import sqlite3
import typing
from datetime import timedelta

import diskcache
import httpx
from packaging.requirements import InvalidRequirement
from packaging.utils import canonicalize_name
from packaging.version import Version
from simple_repository import SimpleRepository, model
from simple_repository.errors import PackageNotFoundError

from . import fetch_projects
from .fetch_description import PackageInfo, package_info
from .short_release_info import ReleaseInfoModel, ShortReleaseInfo


class Crawler:
    """
    A crawler designed to populate and periodically reindex
    the content of the project's metadata database.
    """
    def __init__(
        self,
        http_client: httpx.AsyncClient,
        crawl_popular_projects: bool,
        source: SimpleRepository,
        projects_db: sqlite3.Connection,
        cache: diskcache.Cache,
        reindex_frequency: timedelta = timedelta(days=1),
        release_info_model: typing.Type[ReleaseInfoModel] = ReleaseInfoModel,
    ) -> None:
        self.frequency_seconds = reindex_frequency.total_seconds()
        self._http_client = http_client
        self._source = source
        self._projects_db = projects_db
        self._cache = cache
        self._crawl_popular_projects = crawl_popular_projects
        if os.environ.get("DISABLE_REPOSITORY_INDEXING") != "1":
            self._task = asyncio.create_task(self.run_reindex_periodically())
        self._release_info_model = release_info_model

    async def crawl_recursively(
        self,
        normalized_project_names_to_crawl: typing.Set[str],
    ) -> None:
        """
        Crawl the matadata of the packages in
        normalized_project_names_to_crawl and
        of their dependencies.
        """
        seen: set = set()
        packages_for_reindexing = set(normalized_project_names_to_crawl)
        while packages_for_reindexing - seen:
            remaining_packages = packages_for_reindexing - seen
            pkg_name = remaining_packages.pop()
            logging.info(
                f"Index iteration loop. Looking at {pkg_name}, with {len(remaining_packages)} remaining ({len(seen)} having been completed)",
            )
            seen.add(pkg_name)
            try:
                prj = await self._source.get_project_page(pkg_name)
            except PackageNotFoundError:
                # faulthandler
                continue

            if not prj.files:
                # The project doesn't have any files.
                continue

            releases, latest_version = self._release_info_model.release_infos(prj)

            if latest_version.is_devrelease or latest_version.is_prerelease:
                # Don't bother fetching pre-release only projects.
                continue

            try:
                file, pkg_info = await self.fetch_pkg_info(
                    prj=prj,
                    version=latest_version,
                    releases=releases,
                    force_recache=False,
                )
            except InvalidRequirement as err:
                # See https://discuss.python.org/t/pip-supporting-non-pep508-dependency-specifiers/23107.
                logging.warning(f"Problem handling package {pkg_name}: {err}")
                continue

            for dist in pkg_info.requires_dist:
                dep_name = dist.name
                packages_for_reindexing.add(canonicalize_name(dep_name))

            # Don't DOS the service, we aren't in a rush here.
            await asyncio.sleep(0.01)

    async def refetch_hook(self) -> None:
        # A hook, that can take as long as it likes to execute (asynchronously), which
        # gets called when the periodic reindexing occurs.
        # We periodically want to refresh the project database to make sure we are up-to-date.
        await fetch_projects.fully_populate_db(
            connection=self._projects_db,
            index=self._source,
        )
        packages_w_dist_info = set()
        for cache_type, name, version in self._cache:
            if cache_type == 'pkg-info':
                packages_w_dist_info.add(name)

        popular_projects = []
        if self._crawl_popular_projects:
            # Add the top 100 packages (and their dependencies) to the index
            URL = 'https://hugovk.github.io/top-pypi-packages/top-pypi-packages-30-days.min.json'
            try:
                resp = await self._http_client.get(URL)
                s = resp.json()
                for _, row in zip(range(100), s['rows']):
                    popular_projects.append(row['project'])
            except Exception as err:
                logging.warning(f'Problem fetching popular projects ({err})')
                pass

        await self.crawl_recursively(packages_w_dist_info | set(popular_projects))

    async def run_reindex_periodically(self) -> None:
        logging.info("Starting the reindexing loop")
        while True:
            try:
                await self.refetch_hook()
            except Exception as err:
                logging.exception(err, exc_info=True)
            await asyncio.sleep(self.frequency_seconds)

    # TODO: Refactor this function.
    async def fetch_pkg_info(
        self,
        prj: model.ProjectDetail,
        version: Version,
        releases: dict[Version, ShortReleaseInfo],
        force_recache: bool,
    ) -> tuple[model.File, PackageInfo]:

        key = ('pkg-info', prj.name, str(version))
        if key in self._cache and not force_recache:
            info_file, files_used_for_cache, pkg_info = self._cache[key]
            # Validate that the cached result covers all of the files, and that no new
            # files have been added since the cache was made. In that case, we re-cache.
            if all(
                [
                    file.filename in files_used_for_cache
                    for file in releases[version].files
                ],
            ):
                return info_file, pkg_info

        if force_recache:
            logging.info('Recaching')

        fetch_projects.insert_if_missing(
            self._projects_db,
            canonicalize_name(prj.name),
            prj.name,
        )

        info_file, pkg_info = await package_info(releases[version].files, self._source, prj.name)
        if pkg_info is not None:
            await self.release_info_retrieved(prj, pkg_info)

        self._cache[key] = info_file, releases[version].files, pkg_info
        release_info = releases[version]
        if release_info.is_latest:
            fetch_projects.update_summary(
                self._projects_db,
                name=canonicalize_name(prj.name),
                summary=pkg_info.summary,
                release_date=info_file.upload_time,
                release_version=str(version),
            )

        return info_file, pkg_info

    # TODO: Document this function, or remove it.
    async def release_info_retrieved(self, project: model.ProjectDetail, package_info: PackageInfo) -> None:
        pass



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

"""
Documentation for the simple_repository_browser package

"""
from . import _version  # type: ignore

__version__ = _version.version  # type: ignore



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import datetime
import logging

from simple_repository import SimpleRepository


def create_table(connection):
    con = connection
    with con as cursor:
        cursor.execute(
            '''CREATE TABLE IF NOT EXISTS projects
            (canonical_name text unique, preferred_name text, summary text, release_date timestamp, release_version text)
            ''',
        )


def insert_if_missing(connection, canonical_name, preferred_name):
    with connection as cursor:
        cursor.execute(
            "INSERT OR IGNORE into projects(canonical_name, preferred_name) values (?, ?)",
            (canonical_name, preferred_name),
        )


def remove_if_found(connection, canonical_name):
    with connection as cursor:
        cursor.execute('DELETE FROM projects where canonical_name = ?;', (canonical_name,))


def update_summary(conn, name: str, summary: str, release_date: datetime.datetime, release_version: str):
    with conn as cursor:
        cursor.execute(
            '''
        UPDATE projects
        SET summary = ?, release_date = ?, release_version = ?
        WHERE canonical_name == ?;
        ''', (summary, release_date, release_version, name),
        )


async def fully_populate_db(connection, index: SimpleRepository):
    con = connection
    logging.info('Fetching names from index')

    project_list = await index.get_project_list()
    project_names = [
        (project.normalized_name, project.name) for project in project_list.projects
    ]
    logging.info('Inserting all new names (if any)')
    with con as cursor:
        for canonical_name, name in project_names:
            cursor.execute(
                "INSERT OR IGNORE into projects(canonical_name, preferred_name) values (?, ?)",
                (canonical_name, name),
            )

    with con as cursor:
        db_canonical_names = {row[0] for row in cursor.execute("SELECT canonical_name FROM projects").fetchall()}

    index_canonical_names = {normed_name for normed_name, _ in project_names}

    if not index_canonical_names:
        logging.warning("No names found on the index. Not removing from the database, as this is likely a problem with the index.")
        return

    names_in_db_no_longer_in_index = db_canonical_names - index_canonical_names
    if names_in_db_no_longer_in_index:
        logging.warning(
            f'Removing the following { len(names_in_db_no_longer_in_index) } names from the database:\n   '
            "\n   ".join(list(names_in_db_no_longer_in_index)[:2000]),
        )
    with con as cursor:
        for name in names_in_db_no_longer_in_index:
            cursor.execute(
                '''
                DELETE FROM projects
                WHERE canonical_name == ?;
                ''',
                (name,),
            )
    logging.info('DB synchronised with index')



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import typing


class RequestError(Exception):
    def __init__(
        self,
        status_code: int,
        detail: str,
        *args: typing.Any,
        **kwags: typing.Any,
    ) -> None:
        super().__init__(*args, **kwags)
        self.status_code = status_code
        self.detail = detail


class InvalidSearchQuery(ValueError):
    def __init__(self, msg) -> None:
        super().__init__(msg)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from packaging.requirements import Requirement

from simple_repository_browser.fetch_description import RequirementsSequence


def test_extra__basic():
    s = RequirementsSequence((Requirement('foo'), Requirement('bar; extra == "bar"')))
    assert s.extras() == {'bar'}


def test_extras__multiple_specs():
    s = RequirementsSequence(
        (
            Requirement('bar; extra == "bar" and extra == "foo"'),
            Requirement('wobble; extra == "wibble"'),
        ),
    )
    assert s.extras() == {'bar', 'foo', 'wibble'}


def test_extras__2_extras_or():
    s = RequirementsSequence(
        (
            Requirement('foo'),
            Requirement('bar; extra == "bar" or extra == "foo"'),
        ),
    )
    assert s.extras() == {'bar', 'foo'}


def test_extras__2_extras_and():
    # Not realistic, but technically possible.
    s = RequirementsSequence((Requirement('bar; extra == "bar" and extra == "foo"'),))
    assert s.extras() == {'bar', 'foo'}


def test_extras__and_py_version():
    s = RequirementsSequence(
        (
            Requirement('foo'),
            Requirement('bar; extra == "bar" and python_version>="4.0"'),
        ),
    )
    assert s.extras() == {'bar'}


def test_extras__none():
    s = RequirementsSequence((
        Requirement("foo; (os_name == 'nt' or sys_platform == 'linux') and python_version <= '3.8'"),
        Requirement("bar; python_version > '3.8'"),
    ))
    assert s.extras() == set()



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import itertools
import math
import sqlite3
import typing

import diskcache
from packaging.utils import canonicalize_name
from packaging.version import Version
from simple_repository import SimpleRepository
from simple_repository.errors import PackageNotFoundError
from simple_repository.model import File, ProjectDetail

from . import _search, compatibility_matrix, crawler, errors, fetch_projects
from .fetch_description import PackageInfo
from .short_release_info import ReleaseInfoModel, ShortReleaseInfo


class RepositoryStatsModel(typing.TypedDict):
    n_packages: int
    n_dist_info: int
    n_packages_w_dist_info: int


class QueryResultModel(typing.TypedDict):
    exact: tuple[str, str, str, str] | None
    search_query: str
    results: list[tuple[str, str, str, str]]
    results_count: int  # May be more than in the results list (since paginated).
    single_name_proposal: str | None
    page: int  # Note: starts at 1.
    n_pages: int


class ProjectPageModel(typing.TypedDict):
    # The project detail contents for this project.
    project: ProjectDetail

    # The list of versions for this project. Sorted by version.
    releases: tuple[ShortReleaseInfo, ...]

    # This version.
    this_release: ShortReleaseInfo

    # Classifiers, grouped by the first part of the classifier
    classifiers_by_top_level: dict[str, tuple[str, ...]]

    # The latest stable version of this project.
    latest_release: ShortReleaseInfo

    # The file, found in the project detail page, for which the metadata applies.
    file_info: File

    # The pkg-info metadata, in dictionary form.
    file_metadata: PackageInfo

    # Information about the wheel compatibility
    compatibility_matrix: compatibility_matrix.CompatibilityMatrixModel


class ErrorModel(typing.TypedDict):
    detail: str


class Model:
    def __init__(
        self,
        source: SimpleRepository,
        projects_db: sqlite3.Connection,
        cache: diskcache.Cache,
        crawler: crawler.Crawler,
        release_info_model: typing.Type[ReleaseInfoModel] = ReleaseInfoModel,
    ) -> None:
        self.projects_db = projects_db
        self.source = source
        self.cache = cache
        self.crawler = crawler
        self._release_info_model = release_info_model

    def repository_stats(self) -> RepositoryStatsModel:
        with self.projects_db as cursor:
            [n_packages] = cursor.execute('SELECT COUNT(canonical_name) FROM projects').fetchone()

        with self.cache as cache:
            n_dist_info = len(cache)
            packages_w_dist_info = set()
            for cache_type, name, version in cache:
                if cache_type == 'pkg-info':
                    packages_w_dist_info.add(name)
            n_packages_w_dist_info = len(packages_w_dist_info)

        return RepositoryStatsModel(
            n_packages=n_packages,
            n_dist_info=n_dist_info,
            n_packages_w_dist_info=n_packages_w_dist_info,
        )

    def _compatibility_matrix(self, files: tuple[File, ...]) -> compatibility_matrix.CompatibilityMatrixModel:
        # Compute the compatibility matrix for the given files.
        return compatibility_matrix.compatibility_matrix(files)

    def project_query(self, query: str, page_size: int, page: int) -> QueryResultModel:
        try:
            search_terms = _search.parse(query)
        except _search.ParseError:
            raise errors.InvalidSearchQuery("Invalid search pattern")

        if not search_terms:
            raise errors.InvalidSearchQuery("Please specify a search query")
        try:
            condition_query, condition_terms = _search.build_sql(search_terms)
        except ValueError as err:
            raise errors.InvalidSearchQuery(f"Search query invalid ({str(err)})")

        single_name_proposal = _search.simple_name_from_query(search_terms)
        exact = None

        offset = (page-1) * page_size  # page is 1 based.

        with self.projects_db as cursor:
            result_count = cursor.execute(
                "SELECT COUNT(*) as count FROM projects WHERE "
                f"{condition_query}", condition_terms,
            ).fetchone()
            n_results = result_count['count']

            n_pages = math.ceil(n_results / page_size)
            if n_pages > 0 and (page < 1 or page > n_pages):
                raise errors.InvalidSearchQuery(
                    f"Requested page (page: {page}) is beyond the number of pages ({n_pages})",
                )

            if single_name_proposal:
                exact = cursor.execute(
                    'SELECT canonical_name, summary, release_version, release_date FROM projects WHERE canonical_name == ?',
                    (single_name_proposal,),
                ).fetchone()
            results = cursor.execute(
                "SELECT canonical_name, summary, release_version, release_date FROM projects WHERE "
                f"{condition_query} LIMIT ? OFFSET ?",
                condition_terms + (page_size, offset),
            ).fetchall()

        # Drop the duplicate.
        if exact in results:
            results.remove(exact)

        return QueryResultModel(
            exact=exact,
            search_query=query,
            results=results,
            results_count=n_results,
            single_name_proposal=single_name_proposal,
            page=page,
            n_pages=n_pages,
        )

    async def project_page(
        self,
        project_name: str,
        version: Version | None,
        recache: bool,
    ) -> ProjectPageModel:
        canonical_name = canonicalize_name(project_name)
        try:
            prj = await self.source.get_project_page(canonical_name)
            fetch_projects.insert_if_missing(self.projects_db, canonical_name, project_name)
        except PackageNotFoundError:
            # Tidy up the cache if the project is no longer found.
            for key in list(self.cache):
                if key[:2] == ('pkg-info', canonical_name):
                    self.cache.pop(key)
            fetch_projects.remove_if_found(self.projects_db, canonical_name)
            raise errors.RequestError(status_code=404, detail=f"Project {project_name} not found.")

        if not prj.files:
            raise errors.RequestError(status_code=404, detail=f"No releases for {project_name}.")

        releases, latest_version = self._release_info_model.release_infos(prj)

        if version is None:
            version = latest_version

        if version not in releases:
            raise errors.RequestError(status_code=404, detail=f'Release "{version}" not found for {project_name}.')

        info_file, pkg_info = await self.crawler.fetch_pkg_info(prj, version, releases, force_recache=recache)
        classifiers_by_top_level = {
            top_level: tuple(classifier) for top_level, classifier in itertools.groupby(
                pkg_info.classifiers, key=lambda s: s.split('::')[0],
            )
        }

        compat_mtx = self._compatibility_matrix(releases[version].files)

        return ProjectPageModel(
            project=prj,
            releases=tuple(releases.values()),
            this_release=releases[version],
            classifiers_by_top_level=classifiers_by_top_level,
            latest_release=releases[latest_version],  # Note: May be the same release.
            file_info=info_file,
            file_metadata=pkg_info,
            compatibility_matrix=compat_mtx,
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import dataclasses
import typing
from datetime import datetime

from packaging.utils import canonicalize_name
from packaging.version import InvalidVersion, Version
from simple_repository import model
from simple_repository.packaging import extract_package_version


@dataclasses.dataclass(frozen=True)
class ShortReleaseInfo:
    # A short representation of a release. Intended to be lightweight to compute,
    # such that many ShortReleaseInfo instances can be provided to a view.
    version: Version
    files: tuple[model.File, ...]
    release_date: datetime | None
    is_latest: bool
    yank_status: bool | typing.Literal['partial']


class ReleaseInfoModel:
    @classmethod
    def release_infos(cls, project_detail: model.ProjectDetail) -> tuple[dict[Version, ShortReleaseInfo], Version]:
        files_groued_by_version: dict[Version, list[model.File]] = {}

        if not project_detail.files:
            raise ValueError("No files for the release")

        canonical_name = canonicalize_name(project_detail.name)
        for file in project_detail.files:
            try:
                release = Version(
                    version=extract_package_version(
                        filename=file.filename,
                        project_name=canonical_name,
                    ),
                )
            except (ValueError, InvalidVersion):
                release = Version('0.0rc0')
            files_groued_by_version.setdefault(release, []).append(file)

        result: dict[Version, ShortReleaseInfo] = {}

        latest_version = cls.compute_latest_version(tuple(files_groued_by_version.keys()))

        for version, files in sorted(files_groued_by_version.items()):
            upload_times = [file.upload_time for file in files if file.upload_time]
            if upload_times:
                earliest_release_date = min(upload_times)
            else:
                earliest_release_date = None

            yankeds = [bool(file.yanked) for file in files]
            yank_status: bool | typing.Literal['partial'] = False
            if all(yankeds):
                yank_status = True
            elif any(yankeds):
                yank_status = 'partial'

            result[version] = ShortReleaseInfo(
                version=version,
                files=tuple(files),
                release_date=earliest_release_date,
                is_latest=(version == latest_version),
                yank_status=yank_status,
            )

        return result, latest_version

    @classmethod
    def compute_latest_version(cls, versions: tuple[Version, ...]) -> Version:
        # Use the pip logic to determine the latest release. First, pick the greatest non-dev version,
        # and if nothing, fall back to the greatest dev version. If no release is available return None.
        sorted_versions = sorted(versions)
        for version in sorted_versions[::-1]:
            if not version.is_devrelease and not version.is_prerelease:
                return version
        return sorted_versions[-1]



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import parsley
import pytest

from simple_repository_browser import _search
from simple_repository_browser._search import Filter, FilterOn


@pytest.mark.parametrize(
    ["query", "expected_expression_graph"],
    [
        ("", ()),
        pytest.param("some-name", (Filter(FilterOn.name_or_summary, 'some-name'),)),
        pytest.param("some name", (_search.And(Filter(FilterOn.name_or_summary, 'some'), Filter(FilterOn.name_or_summary, 'name')),)),
        pytest.param("som*name", (Filter(FilterOn.name_or_summary, 'som*name'),)),
        pytest.param('"some name"', (Filter(FilterOn.name_or_summary, '"some name"'),)),
        pytest.param('"some-name"', (Filter(FilterOn.name_or_summary, '"some-name"'),)),
        pytest.param('"CASE"', (Filter(FilterOn.name_or_summary, '"CASE"'),)),
        pytest.param('-foo', (_search.Not(Filter(FilterOn.name_or_summary, 'foo')),)),
        pytest.param('-"foo bar"', (_search.Not(Filter(FilterOn.name_or_summary, '"foo bar"')),)),
        pytest.param('-name:"foo bar"', (_search.Not(Filter(FilterOn.name, '"foo bar"')),)),
        pytest.param('name:foo', (Filter(FilterOn.name, 'foo'),)),
        pytest.param(
            'name:foo OR name:bar', (
                _search.Or(
                    Filter(FilterOn.name, 'foo'),
                    Filter(FilterOn.name, 'bar'),
                ),
            ),
        ),
        pytest.param(
            'name:foo AND "fiddle AND sticks"', (
                _search.And(
                    Filter(FilterOn.name, 'foo'),
                    Filter(FilterOn.name_or_summary, '"fiddle AND sticks"'),
                ),
            ),
        ),
        pytest.param('summary:foo', (Filter(FilterOn.summary, 'foo'),)),
        pytest.param(
            'name:"NAME OR" AND "fiddle AND sticks"', (
                _search.And(
                    Filter(FilterOn.name, '"NAME OR"'),
                    Filter(FilterOn.name_or_summary, '"fiddle AND sticks"'),
                ),
            ),
        ),
        pytest.param('(((a)))', (Filter(FilterOn.name_or_summary, 'a'),)),
        pytest.param('(((a) OR (b)))', (_search.Or(Filter(FilterOn.name_or_summary, 'a'), Filter(FilterOn.name_or_summary, 'b')),)),
        pytest.param(
            '(a AND b) OR (c AND d)', (
                _search.Or(
                    _search.And(Filter(FilterOn.name_or_summary, 'a'), Filter(FilterOn.name_or_summary, 'b')),
                    _search.And(Filter(FilterOn.name_or_summary, 'c'), Filter(FilterOn.name_or_summary, 'd')),
                ),
            ),
        ),
        pytest.param(
            '((a AND b)) OR (c AND -d)', (
                _search.Or(
                    _search.And(Filter(FilterOn.name_or_summary, 'a'), Filter(FilterOn.name_or_summary, 'b')),
                    _search.And(Filter(FilterOn.name_or_summary, 'c'), _search.Not(Filter(FilterOn.name_or_summary, 'd'))),
                ),
            ),
        ),
    ],
)
def test_parse_query(query, expected_expression_graph):
    result = _search.parse(query)
    assert result == expected_expression_graph


@pytest.mark.parametrize(
    ["query", "expected_result"],
    [
        ("", None),
        ("name:foo", "foo"),
        ("name:foo__unnormed", "foo-unnormed"),
        ("foo", "foo"),
        ("some*.Name", None),
        ("summary:\"Some Description\"", None),
        ("foo bar", None),
        ("foo OR bar", None),
        ("-name:foo OR -bar", None),
    ],
)
def test_simple_name_proposal(query, expected_result):
    terms = _search.parse(query)
    result = _search.simple_name_from_query(terms)
    assert result == expected_result


@pytest.mark.parametrize(
    ["query", "expected_predicate"],
    [
        ("", ("", ())),
        ("name:foo", ('canonical_name LIKE ?', ('%foo%',))),
        ("name:foo__unnormed", ('canonical_name LIKE ?', ('%foo-unnormed%',))),
        ("foo", ('(canonical_name LIKE ? OR summary LIKE ?)', ('%foo%', '%foo%'))),
        ("some*.Name", ('(canonical_name LIKE ? OR summary LIKE ?)', ('%some%-name%', '%some%.Name%'))),
        ("summary:\"Some Description\"", ('summary LIKE ?', ('%Some Description%',))),
        ("foo bar", ('((canonical_name LIKE ? OR summary LIKE ?) AND (canonical_name LIKE ? OR summary LIKE ?))', ('%foo%', '%foo%', '%bar%', '%bar%'))),
        ("foo OR bar", ('((canonical_name LIKE ? OR summary LIKE ?) OR (canonical_name LIKE ? OR summary LIKE ?))', ('%foo%', '%foo%', '%bar%', '%bar%'))),
        ("-name:foo OR -bar", ('(Not (canonical_name LIKE ? OR (Not (canonical_name LIKE ? OR summary LIKE ?))))', ('%foo%', '%bar%', '%bar%'))),
        ("summary:\"Some'; DROP TABLE gotcha; ' Description\"", ('summary LIKE ?', ("%Some'; DROP TABLE gotcha; ' Description%",))),
    ],
)
def test_build_sql_predicate(query, expected_predicate):
    sql_stmt, params = _search.query_to_sql(query)
    assert (sql_stmt, params) == expected_predicate
    assert sql_stmt == expected_predicate[0]
    assert params == expected_predicate[1]


@pytest.mark.parametrize(
    ["query", "expected_exception"],
    [
        # ("", ()),   ? Should this be an error? Currently explicitly enabled.
        (" ", pytest.raises(parsley.ParseError)),
        ("'s'", pytest.raises(parsley.ParseError)),
        ("\"imbalanced", pytest.raises(parsley.ParseError)),
        ("unacceptable;char", pytest.raises(parsley.ParseError)),
        ("unacceptable%char", pytest.raises(parsley.ParseError)),
        ("-name:(foo OR bar)", pytest.raises(parsley.ParseError)),
        ("name:", pytest.raises(parsley.ParseError)),
        ('notallowed:foo', pytest.raises(parsley.ParseError)),
    ],
)
def test_invalid_query(query, expected_exception):
    with expected_exception:
        result = _search.parse(query)
        print('Result:', result)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import asyncio
import dataclasses
import datetime
import email.parser
import email.policy
import logging
import os.path
import pathlib
import tempfile
import typing

import httpx
import pkginfo
import readme_renderer.markdown
import readme_renderer.rst
import readme_renderer.txt
from packaging.requirements import Requirement
from simple_repository import SimpleRepository, model


@dataclasses.dataclass
class FileInfo:
    #: Size, in bytes, of the compressed file.
    size: int


class RequirementsSequence(tuple[Requirement]):
    def extras(self) -> set[str]:
        # Get all extras found in any of the contained requirements.
        _extras = set()
        for req in iter(self):
            _extras.update(self.extras_for_requirement(req))
        return _extras

    @classmethod
    def extra_for_requirement(cls, requirement: Requirement) -> str | None:
        extras = list(cls.extras_for_requirement(requirement))
        if len(extras) > 1:
            raise ValueError("Not possible from setuptools")
        elif extras:
            return extras[0]
        else:
            return None

    @classmethod
    def discover_extra_markers(cls, ast) -> typing.Generator[str, None, None]:
        # Find all extra parts of the markers.
        if len(ast) == 1:
            # https://github.com/pypa/packaging/blob/09f131b326453f18a217fe34f4f7a77603b545db/src/packaging/markers.py#L75
            ast = ast[0]
        if isinstance(ast, list):
            if isinstance(ast[0], (list, tuple)):
                yield from cls.discover_extra_markers(ast[0])
            if isinstance(ast[2], (list, tuple)):
                yield from cls.discover_extra_markers(ast[2])
        elif isinstance(ast, tuple):
            lhs_v = getattr(ast[0], 'value', None)
            if lhs_v == 'extra':
                yield ast[2].value
            # Note: Technically, it is possible to build a '"foo" == extra' style
            #       marker. We don't bother with it though, since it isn't something
            #       that comes out of setuptools.
        else:
            raise ValueError(f"Unexpected ast component {ast}")

    @classmethod
    def extras_for_requirement(cls, requirement: Requirement) -> set[str]:
        req_marker = requirement.marker
        if req_marker:
            # Access the AST. Not yet a public API, see https://github.com/pypa/packaging/issues/448.
            markers_ast = req_marker._markers
            return set(list(cls.discover_extra_markers(markers_ast)))
        return set()


@dataclasses.dataclass
class PackageInfo:
    """Represents a simplified pkg-info/dist-info metadata, suitable for easy (and safe) use in html templates"""
    summary: str
    description: str  # This is HTML safe (rendered with readme_renderer).
    author: typing.Optional[str] = None
    maintainer: typing.Optional[str] = None
    classifiers: typing.Sequence[str] = ()
    project_urls: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    requires_python: typing.Optional[str] = None
    requires_dist: RequirementsSequence = RequirementsSequence()
    yanked: bool | str = False

    # A mapping of filename to FileInfo. This must only be used for sharing size information,
    # and will be removed once this code moves to a component based repository definition.
    files_info: dict[str, FileInfo] = dataclasses.field(default_factory=dict)


async def fetch_file(url, dest):
    async with httpx.AsyncClient(verify=False) as http_client:
        async with http_client.stream("GET", url) as r:
            try:
                r.raise_for_status()
            except httpx.HTTPError as err:
                raise IOError(f'Unable to fetch file (reason: { str(err) })')
            chunk_size = 1024 * 100
            with open(dest, 'wb') as fd:
                async for chunk in r.aiter_bytes(chunk_size):
                    fd.write(chunk)


class PkgInfoFromFile(pkginfo.Distribution):
    def __init__(self, filename: str):
        self._filename = filename
        self.extractMetadata()

    def read(self):
        content = pathlib.Path(self._filename).read_text()
        return content.encode()


async def package_info(
    release_files: tuple[model.File, ...],
    repository: SimpleRepository,
    project_name: str,
) -> tuple[model.File, PackageInfo]:
    files = sorted(
        release_files,
        key=lambda file: (
            not file.dist_info_metadata,  # Put those with dist info metadata first.
            not file.filename.endswith('.whl'),
            not file.filename.endswith('.tar.gz'),
            not file.filename.endswith('.zip'),
            file.upload_time,  # Distinguish conflicts by picking the earliest one.
        ),
    )

    files_info: typing.Dict[str, FileInfo] = {}

    # Get the size from the repository, if possible.
    for file in files:
        if file.size:
            files_info[file.filename] = FileInfo(
                size=file.size,
            )

    limited_concurrency = asyncio.Semaphore(10)
    # Compute the size of each file.
    # TODO: This should be done as part of the repository component interface.
    async with httpx.AsyncClient(verify=False) as http_client:
        async def semaphored_head(filename: str, url: str):
            async with limited_concurrency:
                headers: dict[str, str] = {}
                return (
                    filename,
                    await http_client.head(url, follow_redirects=True, headers=headers),
                )
        coros = [
            semaphored_head(file.filename, file.url)
            for file in files
            if file.filename not in files_info
        ]
        for coro in asyncio.as_completed(coros):
            filename, response = await coro
            files_info[filename] = FileInfo(
                size=int(response.headers['Content-Length']),
            )

    file = files[0]

    if file.dist_info_metadata:
        resource_name = file.filename + '.metadata'
    else:
        raise ValueError(f"Metadata not available for {file}")

    logging.info(f'Downloading metadata for {file.filename} from {resource_name}')

    with tempfile.NamedTemporaryFile(
            suffix=os.path.splitext(file.filename)[1],
    ) as tmp:
        resource = await repository.get_resource(project_name, resource_name)

        if isinstance(resource, model.TextResource):
            tmp.write(resource.text.encode())
            if not file.upload_time:
                # If the repository doesn't provide information about the upload time, estimate
                # it from the headers of the resource, if they exist.
                if ct := resource.context.get('creation-date'):
                    if isinstance(ct, str):
                        file = dataclasses.replace(file, upload_time=datetime.datetime.fromisoformat(ct))
        elif isinstance(resource, model.HttpResource):
            await fetch_file(resource.url, tmp.name)
        else:
            raise ValueError(f"Unhandled resource type ({type(resource)})")

        tmp.flush()
        tmp.seek(0)
        info = PkgInfoFromFile(tmp.name)
        description = generate_safe_description_html(info)

        # If there is email information, but not a name in the "author" or "maintainer"
        # attribute, extract this information from the first person's email address.
        # Will take something like ``"Ivan" foo@example.com`` and extract the "Ivan" part.
        def extract_usernames(emails):
            names = []
            parsed = email.parser.Parser(policy=email.policy.default).parsestr(
                f'To: {info.author_email}',
            )
            for address in parsed['to'].addresses:
                names.append(address.display_name)
            return ', '.join(names)

        if not info.author and info.author_email:
            info.author = extract_usernames(info.author_email)

        if not info.maintainer and info.maintainer_email:
            info.maintainer = extract_usernames(info.maintainer_email)

        project_urls = {
            url.split(',')[0].strip().title(): url.split(',')[1].strip()
            for url in info.project_urls or []
        }
        # Ensure that a Homepage exists in the project urls
        if info.home_page and 'Homepage' not in project_urls:
            project_urls['Homepage'] = info.home_page

        sorted_urls = {
            name: url for name, url in sorted(
                project_urls.items(),
                key=lambda item: (item[0] != 'Homepage', item[0]),
            )
        }

        pkg = PackageInfo(
            summary=info.summary or '',
            description=description,
            author=info.author,
            maintainer=info.maintainer,
            classifiers=info.classifiers,
            project_urls=sorted_urls,
            requires_python=info.requires_python,
            requires_dist=RequirementsSequence([Requirement(s) for s in info.requires_dist]),
            # We include files info as it is the only way to influence the file.size of
            # all files (for the files list page). In the future, this can be a standalone
            # component.
            files_info=files_info,
        )

        if not file.size:
            # If the repository doesn't provide information about the size take it from
            # the file info that we gathered.
            file = dataclasses.replace(file, size=files_info[file.filename].size)

        return file, pkg


def generate_safe_description_html(package_info: pkginfo.Distribution) -> str:
    # Handle the valid description content types.
    # https://packaging.python.org/specifications/core-metadata
    description_type = package_info.description_content_type or 'text/x-rst'
    raw_description = package_info.description or ''

    # Seen in the wild (internal only: sps-deep-hysteresis-compensation).
    description_type = description_type.replace('\"', '')

    if description_type == 'text/x-rst' or description_type.startswith('text/x-rst;'):
        return readme_renderer.rst.render(raw_description) or ""
    elif description_type == 'text/markdown' or description_type.startswith('text/markdown;'):  # Seen longer form with orjson
        return readme_renderer.markdown.render(raw_description) or ""
    elif description_type == 'text/plain' or description_type.startswith('text/plain;'):  # seen with nbformat
        return readme_renderer.txt.render(raw_description) or ""
    else:
        raise ValueError(f"Unknown readme format {description_type}")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import base64
import datetime
import pathlib
import pickle
import tarfile
import tempfile
import typing
import zipfile
from dataclasses import replace
from pathlib import Path

import httpx
from simple_repository import errors, model, utils
from simple_repository.components.metadata_injector import \
    MetadataInjectorRepository

ResourceHeaders = typing.TypedDict('ResourceHeaders', {'creation-date': datetime.datetime}, total=False)


class MetadataInjector(MetadataInjectorRepository):
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        project_page = await super().get_project_page(project_name, request_context=request_context)

        files_changed = False
        files = []
        for file in project_page.files:
            if file.url and file.filename.endswith(".tar.gz") and not file.dist_info_metadata:
                file = replace(file, dist_info_metadata=True)
                files_changed = True

            if file.url and file.filename.endswith(".whl") and not file.dist_info_metadata:
                file = replace(file, dist_info_metadata=True)
                files_changed = True
            if file.url and file.filename.endswith(".zip") and not file.dist_info_metadata:
                file = replace(file, dist_info_metadata=True)
                files_changed = True
            files.append(file)
        if files_changed:
            project_page = replace(project_page, files=tuple(files))
        return project_page

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:

        try:
            # Attempt to get the resource from upstream.
            return await super().get_resource(project_name, resource_name, request_context=request_context)
        except errors.ResourceUnavailable:
            if not resource_name.endswith(".metadata"):
                # If we tried to get a resource that wasn't a .metadata one, and it failed,
                # propagate the error.
                raise

        # The resource doesn't exist upstream, and looks like a metadata file has been
        # requested. Let's try to fetch the underlying resource and compute the metadata.

        # Give the cache-key a unique name, to avoid colliding with the cache from other components
        # (such as MetadataInjectorRepository)
        cache_key = f'MetadataInjector__{project_name}/{resource_name}'

        # First, let's attempt to get the metadata out of the cache.
        encoded_metadata = await self._cache.get(cache_key)
        if encoded_metadata:
            decoded_metadata = pickle.loads(base64.b64decode(encoded_metadata.encode('ascii')))
            metadata = decoded_metadata['body']
            headers = decoded_metadata['headers']

        else:
            # Get hold of the actual artefact from which we want to extract
            # the metadata.
            # FIXME: We should be calling the root of the repository here.
            #  We can't do that until we have context though (currently only on master).
            resource = await super().get_resource(
                project_name, resource_name.removesuffix(".metadata"),
            )
            if isinstance(resource, model.HttpResource):
                try:
                    metadata, headers = await self.download_metadata(
                        package_name=resource_name.removesuffix(".metadata"),
                        download_url=resource.url,
                        http_client=self._http_client,
                    )
                except ValueError as e:
                    # If we can't get hold of the metadata from the file then raise
                    # a resource unavailable.
                    raise errors.ResourceUnavailable(resource_name) from e
            elif isinstance(resource, model.LocalResource):
                try:
                    metadata, headers = self.metadata_from_package(resource.path)
                except ValueError as e:
                    raise errors.ResourceUnavailable(resource_name) from e
            else:
                raise errors.ResourceUnavailable(
                    resource_name.removesuffix(".metadata"),
                    "Unable to fetch the resource needed to extract the metadata.",
                )

            # Cache the result for a faster response in the future.
            encoded_metadata_bytes = pickle.dumps({'headers': headers, 'body': metadata})
            await self._cache.set(cache_key, base64.b64encode(encoded_metadata_bytes).decode('ascii'))

        result = model.TextResource(text=metadata)
        result.context.update(headers)
        return result

    async def download_metadata(
            self,
            package_name: str,
            download_url: str,
            http_client: httpx.AsyncClient,
    ) -> tuple[str, ResourceHeaders]:
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_path = pathlib.Path(tmpdir) / package_name
            await utils.download_file(download_url, pkg_path, http_client)
            return self.metadata_from_package(pkg_path)

    def metadata_from_package(self, path: Path) -> tuple[str, ResourceHeaders]:
        return get_metadata_from_package(path)


def get_metadata_from_package(package_path: pathlib.Path) -> tuple[str, ResourceHeaders]:
    if package_path.name.endswith('.whl'):
        return get_metadata_from_wheel(package_path)
    elif package_path.name.endswith('.tar.gz'):
        return get_metadata_from_sdist(package_path)
    elif package_path.name.endswith('.zip'):
        return get_metadata_from_zip(package_path)
    raise ValueError("Package provided is not a wheel or an sdist")


def get_metadata_from_wheel(package_path: pathlib.Path) -> tuple[str, ResourceHeaders]:
    package_tokens = package_path.name.split('-')
    if len(package_tokens) < 2:
        raise ValueError(
            f"Filename {package_path.name} is not normalized according to PEP-427",
        )
    name_ver = package_tokens[0] + '-' + package_tokens[1]

    with zipfile.ZipFile(package_path, 'r') as ziparchive:
        meta_zip_path = name_ver + ".dist-info/METADATA"
        try:
            info = ziparchive.getinfo(meta_zip_path)
            return (
                ziparchive.read(meta_zip_path).decode(),
                {'creation-date': datetime.datetime(*info.date_time)},
            )
        except KeyError as e:
            raise errors.InvalidPackageError(
                "Provided wheel doesn't contain a metadata file.",
            ) from e


def get_metadata_from_sdist(package_path: pathlib.Path) -> tuple[str, ResourceHeaders]:
    archive = tarfile.TarFile.open(package_path)
    names = archive.getnames()

    pkg_info_files = [x.split('/') for x in names if 'PKG-INFO' in x]
    ordered_pkg_info = sorted(pkg_info_files, key=lambda pth: -len(pth))

    for path in ordered_pkg_info:
        candidate = '/'.join(path)
        f = archive.extractfile(candidate)
        if f is None:
            continue
        data = f.read().decode()
        if 'Metadata-Version' in data:
            info = archive.getmember(candidate)
            metadata: ResourceHeaders = {'creation-date': datetime.datetime.fromtimestamp(info.mtime)}
            return data, metadata
    raise ValueError(f"No metadata found in {package_path.name}")


def get_metadata_from_zip(package_path: pathlib.Path) -> tuple[str, ResourceHeaders]:
    # Used by pyreadline. (a zipfile)
    with zipfile.ZipFile(package_path) as archive:
        names = archive.namelist()

        pkg_info_files = [x.split('/') for x in names if 'PKG-INFO' in x]
        ordered_pkg_info = sorted(pkg_info_files, key=lambda pth: -len(pth))

        for path in ordered_pkg_info:
            candidate = '/'.join(path)
            f = archive.open(candidate, mode='r')
            if f is None:
                continue
            data = f.read().decode()
            if 'Metadata-Version' in data:
                info = archive.getinfo(candidate)
                metadata: ResourceHeaders = {'creation-date': datetime.datetime(*info.date_time)}
                return data, metadata
        raise ValueError(f"No metadata found in {package_path.name}")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from simple_repository import model

from simple_repository_browser.compatibility_matrix import compatibility_matrix


def test_compat_mtx__no_wheel():
    mtx = compatibility_matrix(
        (model.File("lxml-4.9.3.tar.gz", "", {}),),
    )
    assert mtx.py_and_abi_names == ()
    assert mtx.platform_names == ()
    assert mtx.matrix == {}


def test_compat_mtx__pure_wheel():
    mtx = compatibility_matrix(
        (model.File("cycler-0.12.1-py3-none-any.whl", "", {}),),
    )
    assert mtx.py_and_abi_names == ('Python 3',)
    assert mtx.platform_names == ('any',)
    assert tuple(mtx.matrix.keys()) == (('Python 3', 'any'),)


def test_compat_mtx__invalid_filename():
    files = [
        model.File("fake-4.9.3-madeup27-madeup2_7m-anything_you_like.whl", "", {}),
    ]

    mtx = compatibility_matrix(files)

    assert mtx.py_and_abi_names == ('madeup27 (madeup2_7m)',)
    assert mtx.platform_names == ('anything_you_like',)


def test_compat_mtx__pyqt6():
    files = [
        model.File("PyQt6-6.5.3-cp37-abi3-macosx_10_14_universal2.whl", "", {}),
        model.File("PyQt6-6.5.3-cp37-abi3-manylinux_2_28_x86_64.whl", "", {}),
        model.File("PyQt6-6.5.3-cp37-abi3-win_amd64.whl", "", {}),
        model.File("PyQt6-6.5.3.tar.gz", "", {}),
    ]

    mtx = compatibility_matrix(files)

    assert mtx.py_and_abi_names == ('CPython >=3.7 (abi3)',)
    assert mtx.platform_names == ('macosx_10_14_universal2', 'manylinux_2_28_x86_64', 'win_amd64')

    assert tuple(mtx.matrix.keys()) == (('CPython >=3.7 (abi3)', 'macosx_10_14_universal2'), ('CPython >=3.7 (abi3)', 'manylinux_2_28_x86_64'), ('CPython >=3.7 (abi3)', 'win_amd64'))


def test_compat_mtx__lxml_py2_flags():
    files = [
        model.File("lxml-4.9.3-cp27-cp27m-manylinux_2_5_i686.whl", "", {}),
        model.File("lxml-4.9.3-cp27-cp27mu-manylinux_2_5_i686.whl", "", {}),
        model.File("lxml-4.9.3-cp27-cp27du-manylinux_2_5_i686.manylinux1_i686.whl", "", {}),
    ]

    mtx = compatibility_matrix(files)

    assert mtx.py_and_abi_names == ('CPython 2.7', 'CPython (debug) (wide) 2.7', 'CPython (wide) 2.7')
    assert mtx.platform_names == ('manylinux1_i686', 'manylinux_2_5_i686')

    assert tuple(mtx.matrix.keys()) == (
        ('CPython 2.7', 'manylinux_2_5_i686'),
        ('CPython (wide) 2.7', 'manylinux_2_5_i686'),
        ('CPython (debug) (wide) 2.7', 'manylinux1_i686'),
        ('CPython (debug) (wide) 2.7', 'manylinux_2_5_i686'),
    )


def test_compat_mtx__unexpected_flags():
    files = [
        model.File("lxml-4.9.3-cp27-cp27abd-manylinux_2_5_i686.whl", "", {}),
    ]

    mtx = compatibility_matrix(files)

    assert mtx.py_and_abi_names == ('CPython (debug) (additional flags: ab) 2.7',)


def test_compat_mtx__underscore_version():
    files = [
        model.File("lxml-4.9.3-cp3_99-cp3_99d-manylinux_2_5_i686.whl", "", {}),
    ]

    mtx = compatibility_matrix(files)

    assert mtx.py_and_abi_names == ('CPython (debug) 3.99',)


def test_compat_mtx__lxml_pypy_flags():
    mtx = compatibility_matrix(
        (model.File("lxml-4.9.3-pp310-pypy310_pp73-manylinux_2_28_x86_64.whl", "", {}),),
    )
    assert mtx.py_and_abi_names == ('PyPy 3.10 (pp73)',)
    assert mtx.platform_names == ('manylinux_2_28_x86_64',)


def test_compat_mtx__none_abi():
    mtx = compatibility_matrix(
        (
            model.File(
                "pydantic_core-2.11.0-cp38-cp38-manylinux2014_x86_64.whl", "", {},
            ),
            model.File("pydantic_core-2.11.0-cp38-none-win32.whl", "", {}),
            model.File("pydantic_core-2.11.0-cp38-none-win_amd64.whl", "", {}),
        ),
    )
    assert mtx.py_and_abi_names == ('CPython 3.8',)
    assert mtx.platform_names == ('manylinux2014_x86_64', 'win32', 'win_amd64')


def test_compat_mtx__abi3_many_python():
    # A completely wild but real name. Specifying multiple Python targets makes no sense when
    # using abi3.
    mtx = compatibility_matrix(
        (
            model.File(
                "PyQt6-6.0.2-cp36.cp37.cp38.cp39-abi3-manylinux_2_28_x86_64.whl", "", {},
            ),
        ),
    )
    # In practice, this should be 1 ('py36_abi3', ), but no attempt has been made
    # to honour this (given how unnecessary it is).
    assert len(mtx.py_and_abi_names) == 4



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from pathlib import Path

from setuptools import setup

HERE = Path(__file__).parent.absolute()
with (HERE / 'README.md').open('rt') as fh:
    LONG_DESCRIPTION = fh.read().strip()


REQUIREMENTS: dict = {
    'core': [
        'httpx',
        'aiosqlite',
        'diskcache',
        'docutils',
        'fastapi',
        'importlib_metadata>=6.0',
        'jinja2',
        'markdown',
        'markupsafe',
        'packaging',
        'parsley',
        'pkginfo',
        'readme-renderer[md]',
        'simple-repository>=0.4',
        'uvicorn',
    ],
    'test': [
        'pytest',
    ],
    'dev': [
        'build',
        'pre-commit',
        'types-setuptools',
    ],
}


setup(
    name='simple-repository-browser',

    author="CERN, BE-CSS-SET",
    description=(
        'A web interface to browse and search packages in any simple package '
        'repository (PEP-503), inspired by PyPI / warehouse'
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url="https://github.com/simple-repository/simple-repository-browser",

    packages=['simple_repository_browser'],
    python_requires='~=3.11',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    package_data={
        '': [
            'static/*/*',
            'static/*',
            'templates/*',
            'templates/base/*',
        ],
    },

    install_requires=REQUIREMENTS.pop('core'),
    entry_points={
        'console_scripts': [
            'simple-repository-browser = simple_repository_browser.__main__:main',
        ],
    },
    extras_require={
        **REQUIREMENTS,
        # The 'dev' extra is the union of 'test' and 'doc', with an option
        # to have explicit development dependencies listed.
        'dev': [
            req
            for extra in ['dev', 'test', 'doc']
            for req in REQUIREMENTS.get(extra, [])
        ],
    },
)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import argparse
import logging
import os
import typing
from pathlib import Path

import uvicorn

from . import __version__
from ._app import AppBuilder

here = Path(__file__).absolute().parent


def configure_parser(parser: argparse.ArgumentParser) -> None:
    parser.description = "Run the simple-repository-browser"

    # Define the function to be called to eventually handle the
    # parsed arguments.
    parser.set_defaults(handler=handler)

    parser.add_argument("index_url", type=str, nargs='?', default='https://pypi.org/simple/')
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", type=int, default=8080)
    parser.add_argument("--cache-dir", type=str, default=Path(os.environ.get('XDG_CACHE_DIR', Path.home() / '.cache')) / 'simple-repository-browser')
    parser.add_argument("--url-prefix", type=str, default="")
    parser.add_argument('--no-popular-project-crawl', dest='crawl_popular_projects', action='store_false', default=True)
    parser.add_argument('--templates-dir', default=here / "templates", type=Path)


def handler(args: typing.Any) -> None:
    app = AppBuilder(
        index_url=args.index_url,
        cache_dir=Path(args.cache_dir),
        template_paths=[
            args.templates_dir,
            # Include the base templates so that the given templates directory doesn't have to
            # implement *all* of the templates. This must be at a lower precedence than the given
            # templates path, so that they can be overriden.
            here/"templates"/"base",
            # Include the "base" folder, such that upstream templates can inherit from "base/...".
            here/"templates",
        ],
        static_files_path=here / "static",
        crawl_popular_projects=args.crawl_popular_projects,
        url_prefix=args.url_prefix,
        browser_version=__version__,
    ).create_app()

    uvicorn.run(
        app=app,
        host=args.host,
        port=args.port,
    )


def main():
    parser = argparse.ArgumentParser()
    configure_parser(parser)
    args = parser.parse_args()
    args.handler(args)


if __name__ == '__main__':
    logging.basicConfig(level=logging.WARNING)
    main()
